import operator as python_operator
from collections import deque
from collections.abc import Callable, MutableSequence, Sequence

from gofra.context import ProgramContext
from gofra.lexer import Token, TokenType
from gofra.parser import Operator, OperatorType
from gofra.parser.intrinsics import Intrinsic
from gofra.typecheck.types import GofraType

type BinaryIntFoldPredicate = Callable[[int, int], int]
BINARY_INT_INTRINSIC_FOLD_PREDICATES: dict[Intrinsic, BinaryIntFoldPredicate] = {
    Intrinsic.PLUS: python_operator.add,
    Intrinsic.MINUS: python_operator.sub,
    Intrinsic.MULTIPLY: python_operator.mul,
    Intrinsic.DIVIDE: python_operator.truediv,
    Intrinsic.MODULUS: python_operator.mod,
    Intrinsic.EQUAL: python_operator.eq,
    Intrinsic.NOT_EQUAL: python_operator.ne,
    Intrinsic.LESS_THAN: python_operator.lt,
    Intrinsic.GREATER_THAN: python_operator.gt,
    Intrinsic.LESS_EQUAL_THAN: python_operator.le,
    Intrinsic.GREATER_EQUAL_THAN: python_operator.ge,
}


def optimize_constant_folding(program: ProgramContext) -> None:
    """Optimize given operators so they dont left unfolded into useless operators.

    For example: 2 2 + folds into single push 4
    or operations that results into dropping from stack being eliminated
    """
    # TODO(@kirillzhosul): This should be refactored due to refactoring core.
    return
    for function in (*program.functions.values(), program.entry_point):
        function.source = _fold_operators(function.source)


def _fold_operators(unoptimized: Sequence[Operator]) -> Sequence[Operator]:
    non_optimizable_offset = 2
    if len(unoptimized) < non_optimizable_offset:
        return unoptimized

    optimized = deque(unoptimized[:non_optimizable_offset])
    jump_idx_left_shift = 0

    open_ctx_stack: list[Operator] = [
        operator
        for operator in optimized
        if operator.type in (OperatorType.DO, OperatorType.IF)
    ]

    def shift_unoptimized_operator() -> None:
        op = optimized[-1]
        if op.jumps_to_operator_idx is not None:
            op.jumps_to_operator_idx = max(
                op.jumps_to_operator_idx - int(jump_idx_left_shift),
                0,
            )

    def shift_open_ctx_operator(shifted_left_by: int) -> None:
        assert len(open_ctx_stack) > 0

        ctx_operator = open_ctx_stack.pop()

        assert ctx_operator.jumps_to_operator_idx is not None
        ctx_operator.jumps_to_operator_idx -= shifted_left_by

    for idx, operator in enumerate(unoptimized[2:], start=2):
        if operator.type in (OperatorType.DO, OperatorType.IF):
            open_ctx_stack.append(operator)
            optimized.append(operator)
            continue

        if operator.type == OperatorType.END:
            optimized.append(operator)
            shift_unoptimized_operator()
            shift_open_ctx_operator(shifted_left_by=jump_idx_left_shift)
            continue

        if operator.type != OperatorType.INTRINSIC:
            optimized.append(operator)
            shift_unoptimized_operator()
            continue

        assert isinstance(operator.operand, Intrinsic)

        if operator.operand == Intrinsic.DROP:
            prev_operator = optimized[-1]
            if prev_operator.type == OperatorType.PUSH_INTEGER:
                optimized.pop()
                jump_idx_left_shift += 2
                continue
            if (
                prev_operator.is_syscall()
                and prev_operator.syscall_optimization_omit_result is False
            ):
                prev_operator.syscall_optimization_omit_result = True
                prev_operator.has_optimizations = True
                jump_idx_left_shift += 1
                continue

        if (
            operator.is_syscall()
            and operator.syscall_optimization_injected_args is None
        ):
            args_count = operator.get_syscall_arguments_count()
            for arg_n in range(args_count):
                arg_operator = optimized.pop()
                if arg_operator.type != OperatorType.PUSH_INTEGER:
                    optimized.append(arg_operator)
                    break

                if operator.syscall_optimization_injected_args is None:
                    operator.has_optimizations = True
                    operator.syscall_optimization_injected_args = [
                        None for _ in range(args_count)
                    ]

                assert isinstance(arg_operator.operand, int)
                operator.syscall_optimization_injected_args[arg_n] = (
                    arg_operator.operand
                )
                jump_idx_left_shift += 1
            if operator.syscall_optimization_injected_args:
                operator.syscall_optimization_injected_args = list(
                    reversed(operator.syscall_optimization_injected_args),
                )
            optimized.append(operator)
            continue

        fold_predicate = BINARY_INT_INTRINSIC_FOLD_PREDICATES.get(operator.operand)
        if fold_predicate is None:
            optimized.append(operator)
            shift_unoptimized_operator()
            continue

        was_folded = _fold_binary_integer_math_operator(
            idx,
            optimized,
            unoptimized,
            fold_predicate=fold_predicate,
        )
        if not was_folded:
            optimized.append(operator)
            shift_unoptimized_operator()
            continue

        jump_idx_left_shift += 2
        shift_unoptimized_operator()

    if jump_idx_left_shift > 0:
        return _fold_operators(unoptimized=list(optimized))

    return optimized


def _fold_binary_integer_math_operator(
    idx: int,
    optimized: MutableSequence[Operator],
    unoptimized: Sequence[Operator],
    *,
    fold_predicate: BinaryIntFoldPredicate,
) -> bool:
    operators = [unoptimized[idx - 1], unoptimized[idx - 2]]

    if not (operators[0].type == operators[1].type == OperatorType.PUSH_INTEGER):
        return False

    operands = [operators[0].operand, operators[1].operand]
    assert isinstance(operands[0], int)
    assert isinstance(operands[1], int)

    infer_type = GofraType.INTEGER
    folded_value = fold_predicate(operands[0], operands[1])
    if isinstance(folded_value, bool):
        infer_type = GofraType.BOOLEAN

    folded_value = int(folded_value)

    optimized.pop()
    optimized.pop()

    operator = Operator(
        type=OperatorType.PUSH_INTEGER,
        token=Token(
            type=TokenType.INTEGER,
            text="__optimized__",
            value=folded_value,
            location=unoptimized[idx].token.location,
        ),
        operand=folded_value,
        has_optimizations=True,
        infer_type_after_optimization=infer_type,
    )

    optimized.append(operator)
    return True



from dataclasses import dataclass
from pathlib import Path

from .tokens import TokenLocation


@dataclass(frozen=False)
class LexerState:
    """State for lexical analysis which only required for internal usages."""

    path: Path

    row: int = 0
    col: int = 0

    line: str = ""

    def current_location(self) -> TokenLocation:
        return TokenLocation(
            filepath=self.path,
            line_number=self.row,
            col_number=self.col,
        )



from __future__ import annotations

import sys
from pathlib import Path
from subprocess import CalledProcessError, run

from gofra.assembler import assemble_program
from gofra.cli.ir import emit_ir_into_stdout
from gofra.consts import GOFRA_ENTRY_POINT
from gofra.gofra import process_input_file
from gofra.optimizer import optimize_program
from gofra.typecheck import validate_type_safety

from .arguments import CLIArguments, parse_cli_arguments
from .errors import cli_gofra_error_handler
from .output import cli_message

WARN_ON_CONFUSING_PROG = False


def cli_entry_point(prog: str | None = None) -> None:
    """CLI main entry."""
    prog = prog if prog else Path(sys.argv[0]).name
    with cli_gofra_error_handler():
        if prog == "__main__.py" and WARN_ON_CONFUSING_PROG:
            cli_message("WARNING", "Running with prog == '__main__.py'")
        args = parse_cli_arguments(prog)

        assert len(args.source_filepaths) == 1

        cli_process_toolchain_on_input_files(args)

        if args.execute_after_compilation:
            if args.output_format != "executable":
                cli_message(
                    level="ERROR",
                    text="Cannot execute after compilation due to output format is not set to an executable!",
                )
                sys.exit(1)
            cli_execute_after_compilation(args)


def cli_process_toolchain_on_input_files(args: CLIArguments) -> None:
    """Process full toolchain onto input source files."""
    cli_message(level="INFO", text="Parsing input files...", verbose=args.verbose)
    context = process_input_file(args.source_filepaths[0], args.include_paths)

    if not args.skip_typecheck:
        cli_message(
            level="INFO",
            text="Validating type safety...",
            verbose=args.verbose,
        )
        validate_type_safety(
            functions={**context.functions, GOFRA_ENTRY_POINT: context.entry_point},
        )

    if not args.disable_optimizations:
        cli_message(
            level="INFO",
            text="Applying optimizations...",
            verbose=args.verbose,
        )
        optimize_program(context)

    if args.ir:
        emit_ir_into_stdout(context)
        sys.exit(0)

    cli_message(
        level="INFO",
        text=f"Assemblying final {args.output_format}...",
        verbose=args.verbose,
    )
    assemble_program(
        verbose=args.verbose,
        output_format=args.output_format,
        context=context,
        output=args.output_filepath,
        target=args.target,
        additional_linker_flags=args.linker_flags,
        additional_assembler_flags=args.assembler_flags,
        build_cache_dir=args.build_cache_dir,
        delete_build_cache_after_compilation=args.delete_build_cache,
    )

    cli_message(
        level="INFO",
        text=f"Compiled input file down to {args.output_format} `{args.output_filepath.name}`!",
        verbose=args.verbose,
    )


def cli_execute_after_compilation(args: CLIArguments) -> None:
    """Run executable after compilation if user requested."""
    cli_message(
        "INFO",
        "Trying to execute compiled file due to execute flag...",
        verbose=args.verbose,
    )
    exit_code = 0
    try:
        run(  # noqa: S602
            [args.output_filepath.absolute()],
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
            check=True,
            shell=True,
        )
    except CalledProcessError as e:
        exit_code = e.returncode
    except KeyboardInterrupt:
        cli_message("WARNING", "Execution was interrupted by user!")
        sys.exit(0)

    level = "INFO" if exit_code == 0 else "ERROR"
    cli_message(
        level,
        f"Program finished with exit code {exit_code}!",
        verbose=args.verbose,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, assert_never

from gofra.parser import Operator, OperatorType
from gofra.parser.intrinsics import Intrinsic

from ._context import TypecheckContext
from .exceptions import (
    TypecheckBlockStackMismatchError,
    TypecheckFunctionTypeContractOutViolatedError,
    TypecheckInvalidBinaryMathArithmeticsError,
    TypecheckInvalidPointerArithmeticsError,
)
from .types import GofraType as T

if TYPE_CHECKING:
    from collections.abc import MutableMapping, MutableSequence, Sequence

    from gofra.parser.functions.function import Function


def validate_type_safety(functions: MutableMapping[str, Function]) -> None:
    """Validate type safety of an program by type checking all given functions."""
    for function in functions.values():
        if function.is_externally_defined:
            continue
        validate_function_type_safety(
            function=function,
            global_functions=functions,
        )


def validate_function_type_safety(
    function: Function,
    global_functions: MutableMapping[str, Function],
) -> None:
    """Emulate and validate function type safety inside."""
    emulated_type_stack = emulate_type_stack_for_operators(
        operators=function.source,
        global_functions=global_functions,
        initial_type_stack=list(function.type_contract_in),
        current_function=function,
    )
    if emulated_type_stack != function.type_contract_out:
        raise TypecheckFunctionTypeContractOutViolatedError(
            function=function,
            type_stack=list(emulated_type_stack),
        )


def emulate_type_stack_for_operators(
    operators: Sequence[Operator],
    global_functions: MutableMapping[str, Function],
    initial_type_stack: Sequence[T],
    current_function: Function,
    blocks_idx_shift: int = 0,
) -> MutableSequence[T]:
    """Emulate and return resulting type stack from given operators.

    Functions are provided so calling it will dereference new emulation type stack.
    """
    context = TypecheckContext(emulated_stack_types=list(initial_type_stack))

    idx_max, idx = len(operators), 0
    while idx < idx_max:
        operator, idx = operators[idx], idx + 1
        match operator.type:
            case OperatorType.WHILE | OperatorType.END:
                ...  # Nothing here as there nothing to typecheck
            case OperatorType.DO | OperatorType.IF:
                context.raise_for_arguments(operator, current_function, T.BOOLEAN)

                # Acquire where this block jumps, shift due to emulation layers
                assert operator.jumps_to_operator_idx
                jumps_to_idx = operator.jumps_to_operator_idx - blocks_idx_shift

                assert operators[jumps_to_idx].type == OperatorType.END

                type_stack = emulate_type_stack_for_operators(
                    operators=operators[idx:jumps_to_idx],
                    global_functions=global_functions,
                    initial_type_stack=context.emulated_stack_types[::],
                    blocks_idx_shift=blocks_idx_shift + idx,
                    current_function=current_function,
                )

                if type_stack != context.emulated_stack_types:
                    raise TypecheckBlockStackMismatchError(
                        operator_begin=operator,
                        operator_end=operators[jumps_to_idx],
                        stack_before_block=context.emulated_stack_types,
                        stack_after_block=type_stack,
                    )

                # Skip this part as we typecheck below and acquire type stack
                idx = jumps_to_idx
            case OperatorType.PUSH_STRING:
                context.push_types(T.POINTER, T.INTEGER)
            case OperatorType.PUSH_MEMORY_POINTER:
                context.push_types(T.POINTER)
            case OperatorType.PUSH_INTEGER:
                assert not operator.has_optimizations, "TBD"
                assert not operator.infer_type_after_optimization, "TBD"
                context.push_types(T.INTEGER)
            case OperatorType.FUNCTION_RETURN:
                return context.emulated_stack_types
            case OperatorType.FUNCTION_CALL:
                assert isinstance(operator.operand, str)

                function = global_functions[operator.operand]
                type_contract_in = function.type_contract_in

                if type_contract_in:
                    context.raise_for_arguments(
                        function,
                        current_function,
                        *type_contract_in,
                        operator=operator,
                    )
                context.push_types(*function.type_contract_out)
            case OperatorType.INTRINSIC:
                assert isinstance(operator.operand, Intrinsic)
                match operator.operand:
                    case Intrinsic.INCREMENT | Intrinsic.DECREMENT:
                        context.raise_for_arguments(
                            operator,
                            current_function,
                            T.INTEGER,
                        )
                        context.push_types(T.INTEGER)
                    case Intrinsic.MULTIPLY | Intrinsic.DIVIDE | Intrinsic.MODULUS:
                        # Math arithmetics operates only on integers
                        # so no pointers/booleans/etc are allowed inside these intrinsics

                        context.raise_for_enough_arguments(
                            operator,
                            current_function,
                            required_args=2,
                        )
                        b, a = (
                            context.pop_type_from_stack(),
                            context.pop_type_from_stack(),
                        )

                        if b != T.INTEGER or a != T.INTEGER:
                            raise TypecheckInvalidBinaryMathArithmeticsError(
                                actual_lhs_type=a,
                                actual_rhs_type=b,
                                operator=operator,
                            )

                        context.push_types(T.INTEGER)
                    case Intrinsic.MINUS | Intrinsic.PLUS:
                        context.raise_for_enough_arguments(
                            operator,
                            current_function,
                            required_args=2,
                        )

                        b, a = (
                            context.pop_type_from_stack(),
                            context.pop_type_from_stack(),
                        )

                        if a == T.POINTER:
                            # Pointer arithmetics
                            if b != T.INTEGER:
                                raise TypecheckInvalidPointerArithmeticsError(
                                    actual_lhs_type=a,
                                    actual_rhs_type=b,
                                    operator=operator,
                                )
                            context.push_types(T.POINTER)
                            continue

                        # Integer math
                        context.push_types(b, a)
                        context.raise_for_arguments(
                            operator,
                            current_function,
                            T.INTEGER,
                            T.INTEGER,
                        )
                        context.push_types(T.INTEGER)

                    case Intrinsic.MEMORY_STORE:
                        context.raise_for_arguments(
                            operator,
                            current_function,
                            T.POINTER,
                            T.INTEGER,
                        )
                    case Intrinsic.MEMORY_LOAD:
                        context.raise_for_arguments(
                            operator,
                            current_function,
                            T.POINTER,
                        )
                        context.push_types(T.INTEGER)
                    case Intrinsic.COPY:
                        context.raise_for_enough_arguments(
                            operator,
                            current_function,
                            required_args=1,
                        )

                        argument_type = context.pop_type_from_stack()
                        context.push_types(argument_type, argument_type)
                    case (
                        Intrinsic.EQUAL
                        | Intrinsic.LESS_EQUAL_THAN
                        | Intrinsic.LESS_THAN
                        | Intrinsic.GREATER_EQUAL_THAN
                        | Intrinsic.GREATER_THAN
                        | Intrinsic.NOT_EQUAL
                    ):
                        context.raise_for_arguments(
                            operator,
                            current_function,
                            T.ANY,
                            T.ANY,
                        )
                        context.push_types(T.BOOLEAN)
                    case Intrinsic.DROP:
                        context.raise_for_arguments(operator, current_function, T.ANY)
                    case (
                        Intrinsic.SYSCALL0
                        | Intrinsic.SYSCALL1
                        | Intrinsic.SYSCALL2
                        | Intrinsic.SYSCALL3
                        | Intrinsic.SYSCALL4
                        | Intrinsic.SYSCALL5
                        | Intrinsic.SYSCALL6
                    ):
                        assert not operator.syscall_optimization_injected_args
                        assert not operator.syscall_optimization_omit_result
                        args_count = operator.get_syscall_arguments_count()

                        argument_types = (T.ANY for _ in range(args_count))
                        context.raise_for_arguments(
                            operator,
                            current_function,
                            *argument_types,
                        )
                        context.push_types(T.INTEGER)
                    case Intrinsic.SWAP:
                        context.raise_for_enough_arguments(
                            operator,
                            current_function,
                            required_args=2,
                        )
                        b, a = (
                            context.pop_type_from_stack(),
                            context.pop_type_from_stack(),
                        )
                        context.push_types(b, a)
                    case _:
                        assert_never(operator.operand)
            case _:
                assert_never(operator.type)

    return context.emulated_stack_types



from collections.abc import Iterable

from gofra.exceptions import GofraError


class NoToolkitForAssemblingError(GofraError):
    toolkit_required: Iterable[str]

    def __init__(self, *args: object, toolkit_required: Iterable[str]) -> None:
        super().__init__(*args)
        self.toolkit_required = toolkit_required

    def __repr__(self) -> str:
        return f"Unable to assemble program due to not all toolkit installed, required: {','.join(self.toolkit_required)}"


class UnsupportedBuilderOperatingSystemError(GofraError):
    def __repr__(self) -> str:
        return "You are on unsupported operating system to compile that target"



from gofra.lexer.tokens import Token
from gofra.preprocessor.exceptions import PreprocessorError


class PreprocessorConditionalNoMacroNameError(PreprocessorError):
    def __init__(self, *args: object, conditional_token: Token) -> None:
        super().__init__(*args)
        self.conditional_token = conditional_token

    def __repr__(self) -> str:
        return f"""No macro name specified at preprocessor condition at {self.conditional_token.location}

Expected macro name after conditional block!"""


class PreprocessorConditionalConsumeUntilEndifContextSwitchError(PreprocessorError):
    def __init__(self, *args: object, conditional_token: Token) -> None:
        super().__init__(*args)
        self.conditional_token = conditional_token

    def __repr__(self) -> str:
        return f"""Preprocessor conditional block got context (file) switch while consuming an block at {self.conditional_token.location}.

This is caused probably by an unclosed preprocessor block (e.g no `#endif`).
(Consuming an conditional block should not switch an tokenizer context, as is means current processed file is probably changed)."""



from gofra.exceptions import GofraError


class PreprocessorError(GofraError):
    """General error within preprocessor.

    Should not be used directly as not provides information about error source!
    """

    def __repr__(self) -> str:
        return """General preprocessor error occurred. 

Please open an issue about that undocumented behavior!
"""



"""Input/Output (IO) helpers for filesystem within toolchain.

It is an part of lexer due to toolchain built on top of lexer (as entry point to parser) and opening an file is responsibility of lexer.
"""

from .io import open_source_file, open_source_file_line_stream

__all__ = ("open_source_file", "open_source_file_line_stream")
